# ltsconfig.tcl --
#
#       Querying and modifying simple configuration files with "sections"
#	Modified inifile package from tcllib
#
# Copyright (c) 2007    Led <ledest@gmail.com>
# Copyright (c) 2003    Aaron Faupell <afaupell@users.sourceforge.net>
#

package provide ltsconfig 0.1

namespace eval conffile {
    variable nexthandle; if {![info exists nexthandle]} {set nexthandle 0}
    variable commentchar; if {![info exists commentchar]} {set commentchar \;}
    variable indentsize; if {![info exists indentsize]} {set indentsize 0}
    variable indentchar; if {![info exists indentchar]} {set indentchar " "}
    variable globalsection; if {![info exists globalsection]} {set globalsection "default"}
}

proc ::conffile::open {conffile {mode r+}} {
    variable nexthandle

    if { ![regexp {^(w|r)\+?$} $mode] } {
        error "$mode is not a valid access mode"
    }

    ::set fh conffile$nexthandle
    ::set tmp [::open $conffile $mode]
    fconfigure $tmp -translation crlf

    namespace eval ::conffile::$fh {
        variable data;     array set data     {}
        variable comments; array set comments {}
        variable sections; array set sections {}
    }
    ::set ::conffile::${fh}::channel $tmp
    ::set ::conffile::${fh}::file    [_normalize $conffile]
    ::set ::conffile::${fh}::mode    $mode

    incr nexthandle
    if { [string match "r*" $mode] } {
        _loadfile $fh
    }
    return $fh
}

# close the file and delete all stored info about it
# this does not save any changes. see ::conffile::commit

proc ::conffile::close {fh} {
    _valid_ns $fh
    ::close [::set ::conffile::${fh}::channel]
    namespace delete ::conffile::$fh
}

# write all changes to disk

proc ::conffile::commit {fh} {
    _valid_ns $fh
    namespace eval ::conffile::$fh {
        if {$mode=="r"} {error "cannot write to read-only file"}
        ::close $channel
        ::set channel [::open $file w]
        ::set char $::conffile::commentchar
	::set indent [string repeat $::conffile::indentchar $::conffile::indentsize]
	foreach sec [concat $::conffile::globalsection [lsort -dictionary [lsearch -not -all -inline -exact [array names sections] $::conffile::globalsection]]] {
		if {[info exists comments($sec)]} {
		    puts $channel "[join $comments($sec) "\n"]\n"
		}
		puts $channel "\[$sec\]"
		foreach key [lsort -dictionary [array names data [::conffile::_globescape $sec]\000*]] {
    		    ::set key [lindex [split $key \000] 1]
    		    if {[info exists comments($sec\000$key)]} {
			puts $channel "$indent[join $comments($sec\000$key) "\n"]"
    		    }
		    puts $channel "$indent$key=$data($sec\000$key)"
		}
		puts $channel ""
	}
        catch {unset char sec key indent}
        close $channel
        ::set channel [::open $file r+]
    }
    return
}

# internal command to read in a file
# see open and revert for public commands

proc ::conffile::_loadfile {fh} {
    namespace eval ::conffile::$fh {
        ::set cur {}
        ::set com {}
        set char $::conffile::commentchar
        seek $channel 0 start
        foreach line [split [read $channel] "\n"] {
	    if {[regexp [format {^[[:blank:]]*%s} $char] $line]} {
                #lappend com [string trim [string range $line [string length $char] end]]
		lappend com [string trim $line [string length $char]]
            } elseif { [string match {\[*\]} $line] } {
                ::set cur [string range $line 1 end-1]
                if { $cur == "" } { continue }
                ::set sections($cur) 1
                if { $com != "" } {
                    ::set comments($cur) $com
                    ::set com {}
                }
            } elseif { [string match {*=*} $line] } {
                ::set line [split $line =]
                ::set key [string trim [lindex $line 0]]
                if { $key == "" || $cur == "" } { continue }
                ::set value [string trim [join [lrange $line 1 end] =]]
                if { [regexp "^(\".*\")\s+${char}(.*)$" $value -> 1 2] } {
                    set value $1
                    lappend com $2
                }
                ::set data($cur\000$key) $value
                if { $com != "" } {
                    ::set comments($cur\000$key) $com
                    ::set com {}
                }
            }
        }
        unset char cur com
        catch { unset line key value 1 2 }
    }
}

# internal command to escape glob special characters

proc ::conffile::_globescape {string} {
    return [string map {* \\* ? \\? \\ \\\\ \[ \\\[ \] \\\]} $string]
}

# internal command to check if a section or key is nonexistant

proc ::conffile::_exists {fh sec args} {
    if { ![info exists ::conffile::${fh}::sections($sec)] } {
        error "no such section \"$sec\""
    }
    if { [llength $args] > 0 } {
        ::set key [lindex $args 0]
        if { ![info exists ::conffile::${fh}::data($sec\000$key)] } {
            error "can't read key \"$key\""
        }
    }
}

# internal command to check validity of a handle

if { [package vcompare [package provide Tcl] 8.4] < 0 } {
    proc ::conffile::_normalize {path} {
	return $path
    }
    proc ::conffile::_valid_ns {name} {
	variable ::conffile::${name}::data
	if { ![info exists data] } {
	    error "$name is not an open INI file"
	}
    }
} else {
    proc ::conffile::_normalize {path} {
	file normalize $path
    }
    proc ::conffile::_valid_ns {name} {
	if { ![namespace exists ::conffile::$name] } {
	    error "$name is not an open INI file"
	}
    }
}

# get and set the conffile comment character

proc ::conffile::setcommentchar { {new {}} } {
    if {$new != ""} {
        if {[string length $new] > 1} { error "comment char must be a single character" }
        ::set ::conffile::commentchar $new
    }
    return $::conffile::commentchar
}

proc ::conffile::setglobalsection {S} {
    ::set ::conffile::globalsection $S
}

proc ::conffile::getglobalsection {} {
    return $::conffile::globalsection
}

proc ::conffile::clearglobalsection {} {
    ::set ::conffile::globalsection ""
}

# return all section names

proc ::conffile::sections {fh} {
    _valid_ns $fh
    return [array names ::conffile::${fh}::sections]
}

# return boolean indicating existance of section or key in section

proc ::conffile::exists {fh sec {key {}}} {
    _valid_ns $fh
    if { $key == "" } {
        return [info exists ::conffile::${fh}::sections($sec)]
    }
    return [info exists ::conffile::${fh}::data($sec\000$key)]
}

# return all key names of section
# error if section is nonexistant

proc ::conffile::keys {fh sec} {
    _valid_ns $fh
    _exists $fh $sec
    ::set keys {}
    foreach x [array names ::conffile::${fh}::data [_globescape $sec]\000*] {
        lappend keys [lindex [split $x \000] 1]
    }
    return $keys
}

# return all key value pairs of section
# error if section is nonexistant

proc ::conffile::get {fh sec} {
    _valid_ns $fh
    _exists $fh $sec
    upvar 0 ::conffile::${fh}::data data
    ::set r {}
    foreach x [array names data [_globescape $sec]\000*] {
        lappend r [lindex [split $x \000] 1] $data($x)
    }
    return $r
}

# return the value of a key
# error if key or section is nonexistant

proc ::conffile::value {fh sec key {default {}}} {
    _valid_ns $fh
    if {$default != "" && ![info exists ::conffile::${fh}::sections($sec)]} {
        return $default
    }
    _exists $fh $sec $key
    return [::set ::conffile::${fh}::data($sec\000$key)]
}

# set the value of a key
# new section or key names are created

proc ::conffile::set {fh sec key value} {
    _valid_ns $fh
    ::set sec [string trim $sec]
    ::set key [string trim $key]
    if { $sec == "" || $key == "" } {
        error "section or key may not be empty"
    }
    ::set ::conffile::${fh}::data($sec\000$key) $value
    ::set ::conffile::${fh}::sections($sec) 1
    return $value
}

# delete a key or an entire section
# may delete nonexistant keys and sections

proc ::conffile::delete {fh sec {key {}}} {
    _valid_ns $fh
    if { $key == "" } {
        array unset ::conffile::${fh}::data [_globescape $sec]\000*
        array unset ::conffile::${fh}::sections [_globescape $sec]
    }
    catch {unset ::conffile::${fh}::data($sec\000$key)}
}

# read and set comments for sections and keys
# may comment nonexistant sections and keys

proc ::conffile::comment {fh sec key args} {
    _valid_ns $fh
    upvar 0 ::conffile::${fh}::comments comments
    ::set r $sec
    if { $key != "" } { append r \000$key }
    if { [llength $args] == 0 } {
        if { ![info exists comments($r)] } { return {} }
        return $comments($r)
    }
    if { [llength $args] == 1 && [lindex $args 0] == "" } {
        unset -nocomplain comments($r)
        return {}
    }
    # take care of any embedded newlines
    for {::set i 0} {$i < [llength $args]} {incr i} {
        ::set args [eval [list lreplace $args $i $i] [split [lindex $args $i] \n]]
    }
    eval [list lappend comments($r)] $args
}

# return the physical filename for the handle

proc ::conffile::filename {fh} {
    _valid_ns $fh
    return [::set ::conffile::${fh}::file]
}

# reload the file from disk losing all changes since the last commit

proc ::conffile::revert {fh} {
    _valid_ns $fh
    namespace eval ::conffile::$fh {
        array set data     {}
        array set comments {}
        array set sections {}
    }
    if { ![string match "w*" $mode] } {
        _loadfile $fh
    }
}
